/**
 * @file framework.h
 * @version 1.0
 *
 * @section DESCRIPTION
 * Includes the necessary headers and defines for the maestro library.
 */

#pragma once

#ifdef _WIN32

#define WIN32_LEAN_AND_MEAN  // Exclude rarely-used stuff from Windows headers
// Windows Header Files
#include <windows.h>

#endif
